/*

Integer.cc -- routines for INTEGER data type
 
Copyright (C) 2001/2003 Norbert Mueller, Tom van Diessen
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/


/*
Changelog: (initial version by Tom, all modifications by Norbert)

  2001-08-20  Modifications in the inheritance: 
              The numbers are no longer METRIC_OBJECTs

  2001-07-19  Initial version from Toms thesis

*/

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "iRRAM_core.h"

#define maxslong 2147483647
#define maxlong 2*(maxslong-1) //****** must be unsigned ******

//***************************************************************************************
// Variables for storing the values during limit-calculations
//***************************************************************************************

long ikonvertanzahl=0;
long ikonvertaufrufe=0;
MP_int_type * ikonvert;

//****************************************************************************************
// CONSTRUCTORS
// IMPORTANT: Objects MUST be initialized before being used.
//****************************************************************************************
//****************************************************************************************

inline INTEGER::INTEGER(MP_int_type y)
{
	value = y;
}

//****************************************************************************************
// Constructing INTEGER from scratch
//****************************************************************************************

INTEGER::INTEGER()
{
	MP_int_init(value);
	MP_int_to_INTEGER(0,value);
}

//****************************************************************************************
// Constructing INTEGER from int (standard integer)
//****************************************************************************************

INTEGER::INTEGER(int i)
{
	MP_int_init(value);
	MP_int_to_INTEGER(i,value);
}

//****************************************************************************************
// Constructing INTEGER from long integer (standard)
//****************************************************************************************

INTEGER::INTEGER(long int i)
{
	MP_int_init(value);
	MP_int_to_INTEGER(i,value);
}

//****************************************************************************************
// Constructing INTEGER from (standard) double
//****************************************************************************************

INTEGER::INTEGER(double d)
{
	MP_int_init(value);
	MP_double_to_INTEGER(d,value);
}


//****************************************************************************************
// Constructing INTEGER from character string
// The base is set to 10 (decimal)
//****************************************************************************************

INTEGER::INTEGER(char* s)
{
  
	MP_int_init(value);
	MP_string_to_INTEGER(s,value,10);
}


//****************************************************************************************
// Copy constructor = from existing INTEGER-object
//****************************************************************************************

INTEGER& INTEGER::operator = (const INTEGER& y)
{
  MP_int_duplicate_wo_init(y.value, value);
  return (*this);
}

//****************************************************************************************
// Copy constructor = from (standard) integer
//****************************************************************************************

INTEGER& INTEGER::operator = (int y)
{
	MP_int_type temp;
	MP_int_init(temp);
	MP_int_to_INTEGER(y,temp);
	MP_int_duplicate_wo_init(temp,value);
	MP_int_clear(temp);
	return (*this);
}

//****************************************************************************************
// Standard Destructor
//****************************************************************************************

INTEGER::~INTEGER()
{
  MP_int_clear(value);
  value=NULL;
}


//****************************************************************************************
// Returns the size of INTEGER-objects in number of limbs
//****************************************************************************************

int INTEGER::size()
{
  return MP_int_getsize(value);
}

//****************************************************************************************
// Shifting an INTEGER-value
//****************************************************************************************

#ifdef MP_has_shift

INTEGER scale(const INTEGER& x, long n)
{
	MP_int_type zvalue;
	MP_int_init(zvalue);
	MP_int_shift(x.value,zvalue,n);
	return zvalue;
}

#else

INTEGER scale(const INTEGER& x, long n)
{
	if (n==0) return x;
	if (n==1) return 2*x;
	INTEGER y=1;
	INTEGER xc=2;
	if (n<0) {xc=y/2;n=-n;}
	if (n==1) return xc*x;
	for (long k=n;k>0;k=k/2)
	{
		if (k%2==1) y*=xc;
		if (k==1) break;
		xc*=xc;
	}
	return y*x;
}

#endif



//****************************************************************************************
// Common infix operators for INTEGER
// +-*/^%...
// Last change: 28.01.2001
//****************************************************************************************

//****************************************************************************************
// Power: return  = x ^ n
// Only with positive long integers as exponent
// 1. argument: INTEGER
// 2. argument: unsigned long int
// return value: INTEGER
//****************************************************************************************

INTEGER operator ^ (const INTEGER& x, unsigned long int n)
{
  MP_int_type zvalue;
  MP_int_init(zvalue);
  MP_int_power_i(x.value,n,zvalue);
  return zvalue;
}

//****************************************************************************************
// Power: return = x ^ y
// Works only with positive exponent, otherwise returns 0
// If the exponent is larger than maxlong, it also returns 0
// 1. argument: INTEGER
// 2. argument: INTEGER
// return value: INTEGER
//****************************************************************************************

INTEGER operator ^ (const INTEGER& x, const INTEGER& y)
{
   if (y==0 || x==1) return 1;
   if (y==1) return x;
   if (y<0) return 0;
   //****** too large *****
   if (y>maxslong) return 0;
   return x^unsigned(MP_INTEGER_to_long(y.value));
}


//****************************************************************************************
// Shifter: returns x << n
// shifts n bits to the left
// 1. argument: INTEGER
// 2. argument: long integer
// return value: INTEGER
//****************************************************************************************

INTEGER operator << (const INTEGER& x, long n)
{
	return scale(x,n);
}

//****************************************************************************************
// Shifter: returns x >> n
// shifts n bits to the right
// 1. argument: INTEGER
// 2. argument: long integer
// return value: INTEGER
//****************************************************************************************

INTEGER operator >> (const INTEGER& x, long n)
{
	return scale(x,-n);
}

//****************************************************************************************
// Addition: returns x + y
// Simple infix addition of 2 INTEGER-objects
// 1. argument: INTEGER
// 2. argument: INTEGER
// return value: INTEGER
//****************************************************************************************


INTEGER operator + (const INTEGER& x, const INTEGER& y)
{
	MP_int_type zvalue;
	MP_int_init(zvalue);
	MP_int_add(x.value,y.value,zvalue);
	return zvalue;
}

//****************************************************************************************
// Addition: returns x + y
// Simple infix addition of INTEGER-object with long integer
// 1. argument: INTEGER
// 2. argument: long integer
// return value: INTEGER
//****************************************************************************************


INTEGER operator + (const INTEGER& x, long y)
{
	MP_int_type zvalue;
	MP_int_init(zvalue);
	if(y<0)MP_int_sub_ui(x.value,-y,zvalue);
	else MP_int_add_ui(x.value,y,zvalue);
	return zvalue;
}

//****************************************************************************************
// Addition: returns x + y
// Simple infix addition of INTEGER-object with long integer
// 1. argument: long integer
// 2. argument: INTEGER
// return value: INTEGER
//****************************************************************************************


INTEGER operator + (long x,const INTEGER&  y)
{
	return y+x;
}


//****************************************************************************************
// Addition: returns x += y
// Simple infix addition of 2 INTEGER-objects using left-hand argument 
// as second right-hand argument
// 1. argument: INTEGER
// return value: INTEGER
//****************************************************************************************

INTEGER& operator += (INTEGER& x, const INTEGER& y)
{
	MP_int_type zvalue;
	MP_int_init(zvalue);
	MP_int_add(x.value,y.value,zvalue);
	MP_int_clear(x.value);
	x.value=zvalue;
	return x;
}

//****************************************************************************************
// Subtraction: returns x - y
// Simple infix difference of 2 INTEGER-objects
// 1. argument: INTEGER
// 2. argument: INTEGER
// return value: INTEGER
//****************************************************************************************

INTEGER operator - (const INTEGER& x, const INTEGER& y)
{
	MP_int_type zvalue;
	MP_int_init(zvalue);
	MP_int_sub(x.value,y.value,zvalue);
	return zvalue;
}

//****************************************************************************************
// Subtraction: returns x - y
// Simple infix difference of INTEGER-object and long integer
// 1. argument: INTEGER
// 2. argument: long integer
// return value: INTEGER
//****************************************************************************************

INTEGER operator - (const INTEGER& x, long y)
{
	MP_int_type zvalue;
	MP_int_init(zvalue);
	if(y<0) MP_int_add_ui(x.value,-y,zvalue);
		else MP_int_sub_ui(x.value,y,zvalue);
	return zvalue;
}

//****************************************************************************************
// Subtraction: returns x - y
// Simple infix difference of INTEGER-object and long integer
// 1. argument: long integer
// 2. argument: INTEGER
// return value: INTEGER
//****************************************************************************************

INTEGER operator - (long x, const INTEGER& y)
{
	return -(y-x);
}


//****************************************************************************************
// Negative: returns -x
// changes sign of INTEGER
// 1. argument: INTEGER
// return value: INTEGER
//****************************************************************************************

INTEGER operator - (const INTEGER& x)
{
	MP_int_type zvalue;
	INTEGER y=INTEGER(long(0));
	MP_int_init(zvalue);
	MP_int_sub(y.value,x.value,zvalue);
	return zvalue;
}

//****************************************************************************************
// Multiplication: returns x * y
// Simple infix multiplication of 2 INTEGER-objects
// 1. argument: INTEGER
// 2. argument: INTEGER
// return value: INTEGER
//****************************************************************************************

INTEGER operator * (const INTEGER& x, const INTEGER& y)
{
	MP_int_type zvalue;
	MP_int_init(zvalue);
	MP_int_mul(x.value,y.value,zvalue);
	return zvalue;
}

//****************************************************************************************
// Multiplication: returns x *= y
// Simple infix multiplication of 2 INTEGER-objects using left-hand argument 
// as second right-hand argument
// 1. argument: INTEGER
// return value: INTEGER
//****************************************************************************************

INTEGER& operator *= (INTEGER& x, const INTEGER& y)
{
	MP_int_type zvalue;
	MP_int_init(zvalue);
	MP_int_mul(x.value,y.value,zvalue);
	MP_int_clear(x.value);
	x=zvalue;
	return x;
}

//****************************************************************************************
// Multiplication: returns x * y
// Simple infix multiplication of INTEGER-object with long integer
// 1. argument: INTEGER
// 2. argument: long integer
// return value: INTEGER
//****************************************************************************************

INTEGER operator * (const INTEGER& x, long int y)
{
	MP_int_type zvalue;
	MP_int_init(zvalue);
	MP_int_mul_si(x.value,y,zvalue);
	return(zvalue);
}

//****************************************************************************************
// Multiplication: returns x * y
// Simple infix multiplication of INTEGER-object with long integer
// 1. argument: long integer
// 2. argument: INTEGER
// return value: INTEGER
//****************************************************************************************

INTEGER operator * (long int x, const INTEGER& y)
{
	MP_int_type zvalue;
	MP_int_init(zvalue);
	MP_int_mul_si(y.value,x,zvalue);
	return(zvalue);
}


//****************************************************************************************
// Multiplication: returns x *= y
// Simple infix multiplication of INTEGER-object and long integer 
// using left-hand argument as second right-hand argument
// 1. argument: long integer
// return value: INTEGER
//****************************************************************************************

INTEGER& operator *= (INTEGER& x, long int n)
{
	return x=x*n;
}

//****************************************************************************************
// Division: returns x / y
// Simple infix division of 2 INTEGER-objects
// The division is a truncated division i.e. the quotient is truncated
// 1. argument: INTEGER
// 2. argument: INTEGER
// return value: INTEGER
//****************************************************************************************


INTEGER operator / (const INTEGER& x, const INTEGER& y)
{
	MP_int_type zvalue;
	MP_int_init(zvalue);
	MP_int_div(x.value,y.value,zvalue);
	return zvalue;
}

//****************************************************************************************
// Division: returns x / y
// Simple infix division of INTEGER-object by long integer
// The division is a truncated division i.e. the quotient is truncated
// 1. argument: INTEGER
// 2. argument: long integer
// return value: INTEGER
//****************************************************************************************

INTEGER operator / (const INTEGER& x, long int n)
{
	MP_int_type zvalue;
	MP_int_init(zvalue);
	if(n<0)
		{
		MP_int_div_ui(x.value,-n,zvalue);
		MP_int_mul_si(zvalue,-1,zvalue);
		}
	else MP_int_div_ui(x.value,n,zvalue);
	return zvalue;
}

//****************************************************************************************
// Modulo: returns x % y
// Simple infix modulo division of 2 INTEGER-objects
// The functions returns the (positive) remainder of a truncated division
// 1. argument: INTEGER
// 2. argument: INTEGER
// return value: INTEGER
//****************************************************************************************

INTEGER operator % (const INTEGER& x, const INTEGER& y)
{
	MP_int_type zvalue;
	MP_int_init(zvalue);
	MP_int_modulo(x.value,y.value,zvalue);
	return zvalue;
}
//****************************************************************************************
// Other common operations on INTEGER
// square, sqrt, faculty, scale, shift
// Last change: 28.01.2001
//****************************************************************************************

//****************************************************************************************
// Square: returns x * x    (x^2)
// Returns the square of the given argument
// 1. argument: INTEGER
// return value: INTEGER
//****************************************************************************************

INTEGER square (const INTEGER& x)
{
	MP_int_type zvalue;
	MP_int_init(zvalue);
	MP_int_mul(x.value,x.value,zvalue);
	return zvalue;
}

//****************************************************************************************
// Square-root: returns SQRT(x)
// Returns the truncated square-root of the given argument
// 1. argument: INTEGER
// return value: INTEGER
//****************************************************************************************

INTEGER sqrt (const INTEGER& x)
{
	MP_int_type zvalue;
	MP_int_init(zvalue);
	MP_int_sqrt(x.value, zvalue);
	return zvalue;
}

//****************************************************************************************
// Faculty: x!
// Returns the faculty of the given argument
// 1. argument: long integer
// return value: INTEGER
//****************************************************************************************

INTEGER faculty (long int x)
{
	MP_int_type zvalue;
	MP_int_init(zvalue);
	MP_int_fac(x,zvalue);
	return zvalue;
}

//****************************************************************************************
// Logarithm to base 2
// returns the truncated logarithm with base 2
// 1. argument: INTEGER
// return value: INTEGER
//****************************************************************************************

INTEGER log2 (const INTEGER& x)
{
	return MP_int_log(x.value);
}

//****************************************************************************************
// Absolute value: |x|
// returns the positive part of the object
// 1. argument: INTEGER
// return value: INTEGER
//****************************************************************************************

INTEGER abs (const INTEGER& x)
{
  MP_int_type zvalue;
  MP_int_init(zvalue);
  MP_int_abs(x.value,zvalue);
  return zvalue;
}


//****************************************************************************************
// boolean operators
// Last change: 28.01.2001
//****************************************************************************************

//****************************************************************************************
// Less than: x < y
// returns boolean value 1 if x<y, 0 otherwise
// 1. argument: INTEGER
// 2. argument: INTEGER
// return value: boolean
//****************************************************************************************

bool operator < (const INTEGER& x, const INTEGER& y)
{
  MP_int_type zvalue;
  MP_int_init(zvalue);
  zvalue=(y-x).value;
  bool result=(MP_int_sign(zvalue)==1);
  return result;
}

//****************************************************************************************
// Less or equal than: x <= y
// returns boolean value 1 if x<=y, 0 otherwise
// 1. argument: INTEGER
// 2. argument: INTEGER
// return value: boolean
//****************************************************************************************

bool operator <= (const INTEGER& x, const INTEGER& y)
{
  return !(y<x);
}

//****************************************************************************************
// Greater than: x > y
// returns boolean value 1 if x>y, 0 otherwise
// 1. argument: INTEGER
// 2. argument: INTEGER
// return value: boolean
//****************************************************************************************

bool operator > (const INTEGER& x, const INTEGER& y)
{
  return y<x;
}

//****************************************************************************************
// Greater or equal than: x >= y
// returns boolean value 1 if x>=y, 0 otherwise
// 1. argument: INTEGER
// 2. argument: INTEGER
// return value: boolean
//****************************************************************************************

bool operator >= (const INTEGER& x, const INTEGER& y)
{
  return y<=x;
}

//****************************************************************************************
// Equal to: x == y
// returns boolean value 1 if x==y, 0 otherwise
// 1. argument: INTEGER
// 2. argument: INTEGER
// return value: boolean
//****************************************************************************************

bool operator == (const INTEGER& x, const INTEGER& y)
{
	return MP_int_sign((y-x).value)==0;
}

//****************************************************************************************
// Equal to: x == i
// returns boolean value 1 if x==i, 0 otherwise
// 1. argument: INTEGER
// 2. argument: long integer
// return value: boolean
//****************************************************************************************

bool operator == (const INTEGER& x, long int i)
{
	return x==INTEGER(i);
}

//****************************************************************************************
// Not equal to: x != y
// returns boolean value 1 if x!=y, 0 otherwise
// 1. argument: INTEGER
// 2. argument: INTEGER
// return value: boolean
//****************************************************************************************

bool operator != (const INTEGER& x, const INTEGER& y)
{
	return MP_int_sign((y-x).value)!=0;
}

//****************************************************************************************
// Positive: returns x>0
// returns boolean value 1 if x positive, 0 otherwise
// 1. argument: INTEGER
// return value: boolean
//****************************************************************************************

bool positive (const INTEGER& x)
{
  return ((MP_int_sign(x.value))==1);
}

//****************************************************************************************
// Maximum: max(x,y)
// returns the largest value of x and y
// 1. argument: INTEGER
// 2. argument: INTEGER
// return value: INTEGER
//****************************************************************************************

INTEGER maximum (const INTEGER& x, const INTEGER& y)
{
   return x>y?x:y;
}

//****************************************************************************************
// Minimum: min(x,y)
// returns the smallest value of x and y
// 1. argument: INTEGER
// 2. argument: INTEGER
// return value: INTEGER
//****************************************************************************************


INTEGER minimum (const INTEGER& x, const INTEGER& y)
{
   return -(maximum(-x,-y));
}


//****************************************************************************************
// output functions
//****************************************************************************************



//****************************************************************************************
// rwrite: writes INTEGER to STOUT
// output function for integer 
// 1. argument: INTEGER
//****************************************************************************************


void rwrite(const INTEGER& x)
{
  if (inlimit) 
    {
      DEBUG1(0,"illegal output in continuous section!\n");
      return;
    }
  if (printfausgaben > printfaufrufe)  
    {
      printfaufrufe++;
      return;
    }
  printfausgaben++;
  printfaufrufe++;

  MP_int_printf(x.value);

}

//*****************************************************************************
// Function rwrite
// output function for integer 
// 1. argument: INTEGER
// 2. argument: long integer : width
//*****************************************************************************

void rwrite(const INTEGER& x, const long w)
{
  if (inlimit) 
    {
      DEBUG1(0,"illegal output in continuous section!\n");
      return;
    }
  if (printfausgaben > printfaufrufe)  
    {
      printfaufrufe++;
      return;
    }
  printfausgaben++;
  printfaufrufe++;

  MP_int_writee(x.value,w);    
}


//*****************************************************************************
// Function swrite
// output function for integer 
// 1. argument: INTEGER
// 2. argument: long integer : width
// 	return:   char*
//*****************************************************************************

char* swrite(const INTEGER& x, const long w){
   	return MP_int_swritee(x.value,w);    
}

//*****************************************************************************
// Function swrite
// output function for integer 
// 1. argument: INTEGER
// 2. argument: long integer : width
//*****************************************************************************

char* swrite(const INTEGER& x){
	return MP_int_sprintf(x.value);    
}

