/*

RATIONAL.cc -- routines for RATIONAL data type
 
Copyright (C) 2001/2003 Norbert Mueller, Tom van Diessen
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/


/*
Changelog: (initial version by Tom, all modifications by Norbert)

  2003-08-29  Addition of functions x+i, x-i, x*i, x/i and i+x etc
              for i of simple type "int"

  2001-08-20  Modifications in the inheritance: 
              The numbers are no longer METRIC_OBJECTs

  2001-07-19  Initial version from Tom's diploma thesis

*/

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "iRRAM_core.h"

#define maxslong 2147483647
#define maxlong 2*(maxslong-1) //****** must be unsigned ******

//***************************************************************************************
// Variables for storing the values during limit-calculations
//***************************************************************************************

long rkonvertanzahl=0;
long rkonvertaufrufe=0;
MP_rat_type * rkonvert;

//****************************************************************************************
// CONSTRUCTORS
// IMPORTANT: Objects MUST be initialized before being used.
//****************************************************************************************
//****************************************************************************************

inline RATIONAL::RATIONAL(MP_rat_type y)
{
	value = y;
}

//****************************************************************************************
// Construncting RATIONAL from scratch
//****************************************************************************************

RATIONAL::RATIONAL()
{
	MP_rat_init(value);
	MP_int_to_RATIONAL(0,value);
}

//****************************************************************************************
// Construnctig RATIONAL from one (standard) integer, the denominator is set to 1
//****************************************************************************************

RATIONAL::RATIONAL(int i)
{
	MP_rat_init(value);
	MP_int_to_RATIONAL(i,value);
}

//****************************************************************************************
// Constructing RATIONAL from two (standard) integer values numerator and denominator,
// the resulting quotient will be canonicalized.
//****************************************************************************************

RATIONAL::RATIONAL(int i,int j)
{
	MP_rat_init(value);
	MP_intint_to_RATIONAL(i,j,value);
}

//****************************************************************************************
// Constructing RATIONAL from long integer, the denominator is set to 1.
//****************************************************************************************

RATIONAL::RATIONAL(long int i)
{
	MP_rat_init(value);
	MP_int_to_RATIONAL(i,value);
}

//****************************************************************************************
// Constructing RATIONAL from two long integers as numerator and denominator,
// the resulting quotient will be canonicalized.
//****************************************************************************************

RATIONAL::RATIONAL(long int i,long int j)
{
	MP_rat_init(value);
	MP_intint_to_RATIONAL(i,j,value);
}

//****************************************************************************************
// Constructing RATIONAL from double, the result is NOT rounded
//****************************************************************************************

RATIONAL::RATIONAL(double d)
{
	MP_rat_init(value);
	MP_double_to_RATIONAL(d,value);
}


//****************************************************************************************
// Constructing RATIONAL from string, the string must be in decimal base
//****************************************************************************************

RATIONAL::RATIONAL(char* s)
{
  
	MP_rat_init(value);
	MP_string_to_RATIONAL(s,value);
}

//****************************************************************************************
// Copy constructor
// Types on both sides must be RATIONAL, left side must already be initialized.
//****************************************************************************************

RATIONAL& RATIONAL::operator = (const RATIONAL& y)
{
  MP_rat_duplicate_wo_init(y.value, value);
  return (*this);
}

//****************************************************************************************
// Copy constructor from (standard) integer
// Left side must be initialized
//****************************************************************************************

RATIONAL& RATIONAL::operator = (int y)
{
	MP_rat_type temp;
	MP_rat_init(temp);
	MP_int_to_RATIONAL(y,temp);
	MP_rat_duplicate_wo_init(temp,value);
	MP_rat_clear(temp);
	return (*this);
}

//****************************************************************************************
// Destructor
// Destroying the object and freeing space occupied by it
//****************************************************************************************

RATIONAL::~RATIONAL()
{
  MP_rat_clear(value);
  value=NULL;
}


//****************************************************************************************
// Returns the size of the object in number of limbs
// this function is used only internally
//****************************************************************************************

int RATIONAL::size()
{
  return MP_rat_getsize(value);
}

//****************************************************************************************
// Shift
// defines shifting function if none exists in the backend
//****************************************************************************************

#ifdef MP_has_shift

RATIONAL scale(const RATIONAL& x, long n)
{
	MP_rat_type zvalue;
	MP_rat_init(zvalue);
	MP_rat_shift(x.value,zvalue,n);
	return zvalue;
}

#else

RATIONAL scale(const RATIONAL& x, long n)
{
	if (n==0) return x;
	if (n==1) return 2*x;
	RATIONAL y=1;
	RATIONAL xc=2;
	if (n<0) {xc=y/2;n=-n;}
	if (n==1) return xc*x;
	for (long k=n;k>0;k=k/2)
	{
		if (k%2==1) y*=xc;
		if (k==1) break;
		xc*=xc;
	}
	return y*x;
}

#endif



//****************************************************************************************
// Common infix operators for RATIONAL
// +-*/^%...
//****************************************************************************************
// Power: return  = x ^ n
// Only with positive long integers as exponent
// 1. argument: RATIONAL
// 2. argument: unsigned long int
// return value: RATIONAL
//****************************************************************************************


RATIONAL operator ^ (const RATIONAL& x, long unsigned int n)
{
  MP_rat_type zvalue;
  MP_rat_init(zvalue);
  MP_rat_power(x.value,n,zvalue);
  return zvalue;
}

//****************************************************************************************
// Power: return  = x ^ n
// Exponent may not be larger as max(long integer)
// 1. argument: RATIONAL
// 2. argument: RATIONAL
// return value: RATIONAL
// experimental
//****************************************************************************************


RATIONAL operator ^ (const RATIONAL& x, const RATIONAL& y)
{
	if (y==0 || x==1) return 1;
   	if (y==1) return x;
	if(y>maxlong) return 0; //****** too large exponent
	// return x^MP_INTEGER_to_long(y.value);
	return x;
}

//****************************************************************************************
// Power: return  = x ^ n
// Only with positive long integers as exponent
// 1. argument: RATIONAL
// 2. argument: long integer
// return value: RATIONAL
//****************************************************************************************

//RATIONAL operator ^ (const RATIONAL& x, long n)
//{
//  MP_rat_type zvalue;
//  MP_rat_init(zvalue);
//  MP_rat_power(x.value,n,zvalue);
//  return zvalue;
//}

//****************************************************************************************
// Shift: return = x << n
// Shifts x by n bits to the left
// 1. argument: RATIONAL
// 2. argument: long integer n
// return value: RATIONAL
//****************************************************************************************

RATIONAL operator << (const RATIONAL& x, long n)
{
	MP_rat_type zvalue;
	MP_rat_init(zvalue);
	MP_rat_shift(zvalue,x.value,n);
	return zvalue;
}

//****************************************************************************************
// Shift: return = x >> n
// Shifts x by n bits to the right
// 1. argument: RATIONAL
// 2. argument: long integer n
// return value: RATIONAL
// Comment: Shifts only the numerator!! still to change
//****************************************************************************************

RATIONAL operator >> (const RATIONAL& x, long n)
{
	MP_rat_type zvalue;
	MP_rat_init(zvalue);
	MP_rat_shift(zvalue,x.value,-n);
	return zvalue;
}

//****************************************************************************************
// Addition: returns x + y
// Simple infix addition of 2 RATIONAL-objects
// 1. argument: RATIONAL
// 2. argument: RATIONAL
// return value: RATIONAL
//****************************************************************************************

RATIONAL operator + (const RATIONAL& x, const RATIONAL& y)
{
	MP_rat_type zvalue;
	MP_rat_init(zvalue);
	MP_rat_add(x.value,y.value,zvalue);
	return zvalue;
}

//****************************************************************************************
// Addition: returns x + y
// Simple infix addition of RATIONAL-object with long integer
// 1. argument: RATIONAL
// 2. argument: long integer
// return value: RATIONAL
//****************************************************************************************

RATIONAL operator + (const RATIONAL& x, long y)
{
	MP_rat_type zvalue;
	MP_rat_init(zvalue);
	if(y<0) MP_rat_sub_ui(x.value,-y,zvalue);
	else MP_rat_add_ui(x.value,y,zvalue);
	return zvalue;
}

//****************************************************************************************
// Addition: returns x + y
// Simple infix addition of RATIONAL-object with long integer
// 1. argument: long integer
// 2. argument: RATIONAL
// return value: RATIONAL
//****************************************************************************************

RATIONAL operator + (long x, const RATIONAL& y)
{
	return y+x;
}

RATIONAL operator + (const RATIONAL& x, int y)
{
	return x+long(y);
}

RATIONAL operator + (int y, const RATIONAL& x)
{
	return long(y)+x;
}

//****************************************************************************************
// Addition: returns x += y
// Simple infix addition of 2 RATIONAL-objects using left-hand argument 
// as second right-hand argument
// 1. argument: RATIONAL
// return value: RATIONAL
//****************************************************************************************

RATIONAL& operator += (RATIONAL& x, const RATIONAL& y)
{
	MP_rat_type zvalue;
	MP_rat_init(zvalue);
	MP_rat_add(x.value,y.value,zvalue);
	MP_rat_clear(x.value);
	x.value=zvalue;
	return x;
}

//****************************************************************************************
// Subtraction: returns x - y
// Simple infix difference of 2 RATIONAL-objects
// 1. argument: RATIONAL
// 2. argument: RATIONAL
// return value: RATIONAL
//****************************************************************************************

RATIONAL operator - (const RATIONAL& x, const RATIONAL& y)
{
	MP_rat_type zvalue;
	MP_rat_init(zvalue);
	MP_rat_sub(x.value,y.value,zvalue);
	return zvalue;
}

//****************************************************************************************
// Subtraction: returns x - y
// Simple infix difference of RATIONAL-object and long integer
// 1. argument: RATIONAL
// 2. argument: long integer
// return value: RATIONAL
//****************************************************************************************

RATIONAL operator - (const RATIONAL& x, long y)
{
	MP_rat_type zvalue;
	MP_rat_init(zvalue);
	if(y<0) MP_rat_add_ui(x.value,-y,zvalue);
	else MP_rat_sub_ui(x.value,y,zvalue);
	return zvalue;
}

RATIONAL operator - (const RATIONAL& x, int y)
{
	return x-long(y);
}

RATIONAL operator - (int y, const RATIONAL& x)
{
	return long(y)-x;
}

//****************************************************************************************
// Subtraction: returns x - y
// Simple infix difference of RATIONAL-object and long integer
// 1. argument: long integer
// 2. argument: RATIONAL
// return value: RATIONAL
//****************************************************************************************

RATIONAL operator - (long x, const RATIONAL& y)
{
	return -(y-x);
}


//****************************************************************************************
// Negative: returns -x
// changes sign of RATIONAL
// 1. argument: RATIONAL
// return value: RATIONAL
//****************************************************************************************

RATIONAL operator - (const RATIONAL& x)
{
	MP_rat_type zvalue;
	RATIONAL y=RATIONAL(long(0));
	MP_rat_init(zvalue);
	MP_rat_sub(y.value,x.value,zvalue);
	return zvalue;
}

//****************************************************************************************
// Multiplication: returns x * y
// Simple infix multiplication of 2 RATIONAL-objects
// 1. argument: RATIONAL
// 2. argument: RATIONAL
// return value: RATIONAL
//****************************************************************************************

RATIONAL operator * (const RATIONAL& x, const RATIONAL& y)
{
	MP_rat_type zvalue;
	MP_rat_init(zvalue);
	MP_rat_mul(x.value,y.value,zvalue);
	return zvalue;
}

//****************************************************************************************
// Multiplication: returns x * y
// Simple infix multiplication of RATIONAL-object and long integer
// 1. argument: RATIONAL
// 2. argument: long integer
// return value: RATIONAL
//****************************************************************************************

RATIONAL operator * (const RATIONAL& x, long y)
{
	MP_rat_type zvalue;
	MP_rat_init(zvalue);
	MP_rat_mul_si(x.value,y,zvalue);
	return zvalue;
}

//****************************************************************************************
// Multiplication: returns x * y
// Simple infix multiplication of RATIONAL-object and long integer
// 1. argument: long integer
// 2. argument: RATIONAL
// return value: RATIONAL
//****************************************************************************************

RATIONAL operator * (long x, const RATIONAL& y)
{
	return y*x;
}

RATIONAL operator * (const RATIONAL& x, int y)
{
	return x*long(y);
}

RATIONAL operator * (int y, const RATIONAL& x)
{
	return long(y)*x;
}


//****************************************************************************************
// Multiplication: returns x *= y
// Simple infix multiplication of 2 RATIONAL-objects using left-hand argument 
// as second right-hand argument
// 1. argument: RATIONAL
// return value: RATIONAL
//****************************************************************************************

RATIONAL& operator *= (RATIONAL& x, const RATIONAL& y)
{
	MP_rat_type zvalue;
	MP_rat_init(zvalue);
	MP_rat_mul(x.value,y.value,zvalue);
	MP_rat_clear(x.value);
	x=zvalue;
	return x;
}

//****************************************************************************************
// Multiplication: returns x *= y
// Simple infix multiplication of RATIONAL-object and long integer 
// using left-hand argument as second right-hand argument
// 1. argument: long integer
// return value: RATIONAL
//****************************************************************************************

RATIONAL& operator *= (RATIONAL& x, long n)
{
	return x*=RATIONAL(n);
}


//****************************************************************************************
// Division: returns x / y
// Simple infix division of 2 RATIONAL-objects
// The division is a truncated division i.e. the quotient is truncated
// 1. argument: RATIONAL
// 2. argument: RATIONAL
// return value: RATIONAL
//****************************************************************************************

RATIONAL operator / (const RATIONAL& x, const RATIONAL& y)
{
	MP_rat_type zvalue;
	MP_rat_init(zvalue);
	MP_rat_div(x.value,y.value,zvalue);
	return zvalue;
}

//****************************************************************************************
// Division: returns x / y
// Simple infix division of RATIONAL-object by long integer
// The division is a truncated division i.e. the quotient is truncated
// 1. argument: RATIONAL
// 2. argument: long integer
// return value: RATIONAL
//****************************************************************************************

RATIONAL operator / (const RATIONAL& x, long y)
{
	MP_rat_type zvalue;
	MP_rat_init(zvalue);
	MP_rat_div_si(x.value,y,zvalue);
	return zvalue;
	
}

//****************************************************************************************
// Division: returns x / y
// Simple infix division of RATIONAL-object by long integer
// The division is a truncated division i.e. the quotient is truncated
// 1. argument: RATIONAL
// 2. argument: long integer
// return value: RATIONAL
//****************************************************************************************

RATIONAL operator / (long x, const RATIONAL& y)
{
	MP_rat_type zvalue;
	MP_rat_init(zvalue);
	MP_rat_si_div(x,y.value,zvalue);
	return zvalue;
	
}

RATIONAL operator / (const RATIONAL& x, int y)
{
	return x/long(y);
}

RATIONAL operator / (int y, const RATIONAL& x)
{
	return long(y)/x;
}

//****************************************************************************************
// Other common operations on RATIONAL
// square, sqrt, faculty, scale, shift
// Last change: 28.01.2001
//****************************************************************************************

//****************************************************************************************
// Square: returns x * x    (x^2)
// Returns the square of the given argument
// 1. argument: RATIONAL
// return value: RATIONAL
//****************************************************************************************

RATIONAL square (const RATIONAL& x)
{
	MP_rat_type zvalue;
	MP_rat_init(zvalue);
	MP_rat_mul(x.value,x.value,zvalue);
	return zvalue;
}


//****************************************************************************************
// Logarithm to base 2
// returns the truncated logarithm with base 2
// 1. argument: RATIONAL
// return value: RATIONAL
//****************************************************************************************

RATIONAL log2 (const RATIONAL& x){
	return MP_rat_log(x.value);
}


//****************************************************************************************
// Absolute value: |x|
// returns the positive part of the object
// 1. argument: RATIONAL
// return value: RATIONAL
//****************************************************************************************

RATIONAL abs (const RATIONAL& x){
  MP_rat_type zvalue;
  MP_rat_init(zvalue);
  MP_rat_abs(x.value,zvalue);
  return zvalue;
}

//******************************************************************************************
// boolean operators
// Last change: 28.01.2001
//******************************************************************************************
//****************************************************************************************
// Less than: x < y
// returns boolean value 1 if x<y, 0 otherwise
// 1. argument: RATIONAL
// 2. argument: RATIONAL
// return value: boolean
//****************************************************************************************

bool operator < (const RATIONAL& x, const RATIONAL& y){
  MP_rat_type zvalue;
  MP_rat_init(zvalue);
  zvalue=(y-x).value;
  bool result=(MP_rat_sign(zvalue)==1);
  return result;
}

//****************************************************************************************
// Less or equal than: x <= y
// returns boolean value 1 if x<=y, 0 otherwise
// 1. argument: RATIONAL
// 2. argument: RATIONAL
// return value: boolean
//****************************************************************************************

bool operator <= (const RATIONAL& x, const RATIONAL& y){
  return !(y<x);
}

//****************************************************************************************
// Greater than: x > y
// returns boolean value 1 if x>y, 0 otherwise
// 1. argument: RATIONAL
// 2. argument: RATIONAL
// return value: boolean
//****************************************************************************************

bool operator > (const RATIONAL& x, const RATIONAL& y){
  return y<x;
}

//****************************************************************************************
// Greater or equal than: x >= y
// returns boolean value 1 if x>=y, 0 otherwise
// 1. argument: RATIONAL
// 2. argument: RATIONAL
// return value: boolean
//****************************************************************************************

bool operator >= (const RATIONAL& x, const RATIONAL& y){
  return y<=x;
}

//****************************************************************************************
// Equal to: x == y
// returns boolean value 1 if x==y, 0 otherwise
// 1. argument: RATIONAL
// 2. argument: RATIONAL
// return value: boolean
//****************************************************************************************

bool operator == (const RATIONAL& x, const RATIONAL& y){
	return MP_rat_sign((y-x).value)==0;
}

//****************************************************************************************
// Not equal to: x != y
// returns boolean value 1 if x!=y, 0 otherwise
// 1. Argument: RATIONAL
// 2. Argument: RATIONAL
// return value: boolean
//****************************************************************************************

bool operator != (const RATIONAL& x, const RATIONAL& y){
	return MP_rat_sign((y-x).value)!=0;
}

//****************************************************************************************
// Equal to long: x == i
// returns boolean value 1 if x==i, 0 otherwise
// 1. argument: INTEGER
// 2. argument: long integer
// return value: boolean
//****************************************************************************************

bool operator == (const RATIONAL& x, int i){
	return x==RATIONAL(i);
}

//****************************************************************************************
// Maximum: max(x,y)
// returns the largest value of x and y
// 1. argument: RATIONAL
// 2. argument: RATIONAL
// return value: RATIONAL
//****************************************************************************************

RATIONAL maximum (const RATIONAL& x, const RATIONAL& y){
   return x>y?x:y;
}

//****************************************************************************************
// Minimum: min(x,y)
// returns the smallest value of x and y
// 1. argument: RATIONAL
// 2. argument: RATIONAL
// return value: RATIONAL
//****************************************************************************************

RATIONAL minimum (const RATIONAL& x, const RATIONAL& y){
   return -(maximum(-x,-y));
};




//******************************************************************************************
// output functions
//******************************************************************************************


//****************************************************************************************
// rwrite: writes RATIONAL to STOUT
// 1. argument: RATIONAL
//****************************************************************************************

void rwrite(const RATIONAL& x){
  if (inlimit) 
    {
      DEBUG1(0,"illegal output in continuous section!\n");
      return;
    }
  if (printfausgaben > printfaufrufe)  
    {
      printfaufrufe++;
      return;
    }
  printfausgaben++;
  printfaufrufe++;
  
  MP_rat_writef(x.value,1000);
}

//*****************************************************************************
// Function rwrite
// output function for RATIONAL
//*****************************************************************************

void rwrite(const RATIONAL& x, const long w){
  if (inlimit) 
    {
      DEBUG1(0,"illegal output in continuous section!\n");
      return;
    }

  if (printfausgaben > printfaufrufe)  
    {
      printfaufrufe++;
      return;
    }
  printfausgaben++;
  printfaufrufe++;

  MP_rat_writee(x.value,w);    
}

//*****************************************************************************
// Function swrite
// string function for RATIONAL
// writes rational into string
//*****************************************************************************

char* swrite(const RATIONAL& x, const long w){
  	return MP_rat_swritee(x.value,w);    
}

//****************************************************************************************
// swrite: writes RATIONAL to string
// 1. argument: RATIONAL
//****************************************************************************************

char* swrite(const RATIONAL& x){
	return MP_rat_sprintf(x.value);
}
